/***************
 *  FSHRINK.C
 ***************

Copyright (c) 1998-1999 David Kuehling

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
 *
 * This program creates compressed archives of TI-Graph Link Fargo programs.
 * These archives are stored into TI-Graph Link string variables, using the
 * routines of `STRING92.C'. The compression algorithm is contained in the
 * module `SHRINK.C'.
 *
 * Last modification: january 1999
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <stdarg.h>
#include "shrink.h"
#include "string92.h"


#define VERSION "1.00"

/****************************************************************************/

typedef struct PRGM
{
   unsigned char *data;
   char          *comment,
                 name[9];
   long          len;
} PRGM;

typedef struct PRGM_LIST
{
   PRGM             *prgm;
   struct PRGM_LIST *next;
} PRGM_LIST;

/****************************************************************************/


/**
 * Print an error message in printf - format and exit with exit code 1.
 */
void ErrorExit (char *format, ...)
{
  va_list arg;

  va_start (arg, format);
  fputs ("\nERROR: ", stderr);
  vfprintf (stderr, format, arg);
  va_end (arg);
  fputs ("\n\n", stderr);
  exit (1);
}

/**
 * Add a program `prgm' to the program list `prgmList'. The file is added
 * at the alphabetical correct position
 */
void AddPrgm (PRGM *prgm, PRGM_LIST **prgmList)
{
   PRGM_LIST **entry,            // an entry in the program list
             *newEntry;         // new entry in the program list

   newEntry = malloc (sizeof (PRGM_LIST));
   newEntry->prgm = prgm;

   for (entry = prgmList;; entry = &(*entry)->next)
      if (*entry == NULL || strcmp ((*entry)->prgm->name, prgm->name) > 0)
      {
         newEntry->next = *entry;
         *entry = newEntry;
         break;
      }

}

/**
 * Load a Fargo Program out of a TI Graph Link *.92p file.
 */
PRGM *LoadFargoProgram (char *fileName)
{
   unsigned char  sig[10];     // TI-Graph Link signature
   unsigned short checksum;   // TI-Graph Link file checksum
   PRGM           *prgm;      // loaded fargo program
   FILE           *hdl;

   printf ("Loading `%s' ...\n", fileName);
   if ((hdl = fopen (fileName, "rb")) == NULL)
      ErrorExit (strerror (errno));

   if ((prgm = malloc (sizeof (PRGM))) == NULL)
      ErrorExit ("Not enough memory");

   /** check signature */
   fread (&sig, 1, 10, hdl);
   if (memcmp (sig, "**TI92**\x01\x00", 10))
      ErrorExit ("Not a valid TI-Graph Link file");

   /** read file */
   fseek (hdl, 0x40, SEEK_SET);
   prgm->name[8] = '\0';
   fread (prgm->name, 1, 8, hdl);

   fseek (hdl, 0x4C, SEEK_SET);
   prgm->len = (fgetc (hdl)     | fgetc (hdl)<< 8 |
                fgetc (hdl)<<16 | fgetc (hdl)<<24 ) - 0x58;

   fseek (hdl, 0x56, SEEK_SET);
   if ((prgm->data = malloc (prgm->len)) == NULL)
      ErrorExit ("Not enough memory");
   fread (prgm->data, 1, prgm->len, hdl);

   /** check file and extract comment */
       // Fargo 0.1.x
   if (!memcmp (prgm->data+2, "\x00""P10", 4))
      prgm->comment = (prgm->data+2) + (prgm->data[0xC+2]<<8 | prgm->data[0xD+2]);
       // Fargo II
   else if (!memcmp (prgm->data+2, "\x00""2EXE APPL", 10))
      prgm->comment = prgm->data + (prgm->data[0x16]<<8 | prgm->data[0x17]);
   else
      ErrorExit ("File is neither a Fargo 0.1.x nor a Fargo II program");

   checksum = fgetc (hdl) | fgetc (hdl)<<8;
   if (checksum != GenerateChecksum (prgm->data, prgm->len))
      ErrorExit ("Invalid checksum");

   fclose (hdl);

   printf ("  Comment: `%s'\n", prgm->comment);

   return (prgm);
}

/**
 * Create a program archive, consisting of the programs given by `prgms'
 * If `addComments' == 0: make the poiners to the program comments to point
 * to a zero-string.
 * Archive's format:
 *   $0000.w = size of archive (not including this word)
 *   $0002.l = archive signature ('A' 'R' 'C' #0)
 *   ; pseudo VAT begins here
 *   $0008.w = (this word originally belongs to the VAT but is irrelevant
 *             for the brwselib::SurfList routine) relative pointer to
 *             archive data (relative from $0002)
 *   ; entry of 1st file
 *   $000A.w = number of files in archive
 *   $000C.b [8] = name of first file
 *   $0014.w = relative pointer to comment of file (relative from itself - 8)
 *   $0016.w = number of archive section, this file is stored in
 *   ; entry of 2nd file
 *   .... (same as at $000A)
 *
 */
void CreatePrgmArchive (PRGM_LIST *prgms, char *fileName, int addComments)
{
   unsigned char *prgmData[MAX_FILES],  // compression input data array
                 *arcData,              // compression output data
                 *strData,              // data that are stored in the string
                 *data;                 // current output position in `strData'
   long          prgmLen[MAX_FILES],    // lengths of compressin input data
                 arcLen,                // length of Shrink archive
                 strLen,                // length of string variable
                 totalLen;              // length of all input prgms together
   int           prgmNum,               // number of programs
                 I;
   unsigned      comPtrs[MAX_FILES],    // relative pointers to comments
                 comPtr,                // relative pointer to current comment
                 arcPtr,                // relative pointer to shrink-archive
                 oddArcPtr;             // archive pointer is odd?
   PRGM_LIST     *prgm;                 // pointer to current program in list

   /** convert programs to the right format to be compressed */
   for (prgm = prgms, prgmNum = totalLen = 0; prgm; prgm = prgm->next, prgmNum++)
   {
      prgmData[prgmNum] = prgm->prgm->data;
      totalLen += (prgmLen[prgmNum] = prgm->prgm->len);
   }

   /* compress the programs */
   if ((arcLen = Compress (prgmData, prgmLen, prgmNum, &arcData)) == -1)
   {
      static char *errStr[] =
      {
         "Not enough memory",
         "Too many files",
         "Input file too big",
         "Output file became too big (>65535 bytes)"
      };

      ErrorExit (errStr[ShrinkError-1]);
   }

   /** calculate the relative comment pointers and the string variable's length */
   for (prgm = prgms, comPtr = prgmNum*12, arcPtr = 8, I = 0;
        prgm; prgm = prgm->next, I++)
   {
      if (addComments)
      {
         comPtrs[I] = comPtr;
         arcPtr += strlen (prgm->prgm->comment) + 1 + 12;
         comPtr += strlen (prgm->prgm->comment) + 1 - 12;
      }
      else
      {
         comPtrs[I] = 8;
         arcPtr += 12;
      }
   }

   oddArcPtr = arcPtr&1;        // store whether archive has been aligned
   arcPtr = (arcPtr+1)&0xFFFE;  // align archive
   strLen = arcPtr + arcLen;    // length of string variable

   if (strLen >= 65536)
      ErrorExit ("Outut file became too big (>65535 bytes)");
   if ((strData = malloc (strLen)) == NULL)
      ErrorExit ("Not enough memory");

   /** create the string variable */
   data = strData;

   /* output archive signature */
   *data++ = 'A';
   *data++ = 'R';
   *data++ = 'C';
   *data++ = '\0';
   /* output pseudo-VAT */
   *data++ = arcPtr>>8;
   *data++ = arcPtr&0xFF;
   *data++ = prgmNum>>8;
   *data++ = prgmNum&0xFF;
   for (prgm = prgms, I = 0; prgm; prgm = prgm->next, I++)
   {
      strncpy (data, prgm->prgm->name, 8);
      data += 8;
      *data++ = comPtrs[I]>>8;
      *data++ = comPtrs[I]&0xFF;
      *data++ = I<<8;
      *data++ = I&0xFF;
   }
   /* output comments */
   if (addComments)
      for (prgm = prgms; prgm; prgm = prgm->next)
      {
         strcpy (data, prgm->prgm->comment);
         data += strlen (prgm->prgm->comment)+1;
      }

   /* output archive */
   memcpy (data+oddArcPtr, arcData, arcLen);

   /** output string variable */
   printf ("Creating string variable `%s'...\n", fileName);
   if (Create92S (strData, strLen, fileName))
      ErrorExit ("Can't create string variable `%s' : `%s'",
                 fileName, strerror (errno));

   strLen += 5; // real length (including size and type)
   printf ("  %li bytes in %i files compressed to %li bytes. Ratio: %.2f%%\n",
           totalLen, prgmNum, strLen, (double) (totalLen-strLen) * 100 / totalLen);
}

/**
 * Display a short help
 */
void Help ()
{
   static char *text[] =
   {
      "FSHRINK filename [filename [filename ...]] [-n] [-o...]",
      "  filename   Name of input file, may contain wildcards. Valid input files are",
      "             TI-Graph Link *.92p Fargo 0.1.x or Fargo II programs.",
      "  -o<name>   Specify the name of the archive to create. The program archive",
      "             will be stored into a TI-Graph Link string variable. For this",
      "             reason your file should end on .92p",
      "  -n         Do not include the program's comments into the archive. This will",
      "             save some memory.",
      NULL
   };
   int I;

   for (I = 0; text[I]; I++)
      puts (text[I]);

   exit (0);
}

/**
 * MAIN
 */
int main (int argn, char *argv[])
{
   char          outName[512] = "";   // name of output file
   int           I,
                 addComments = 1;     // add comments to archive ?
   PRGM_LIST     *prgms = NULL;       // list of programs

   puts ("\t----------------------------------------------------------------");
   puts ("\t            Fargo Program Compressor v"VERSION" for FPL II");
   puts ("\t                     by David Khling 1999");
   puts ("\t----------------------------------------------------------------\n");

   /** process arguments */
   for (I = 1; I < argn; I++)
   {
      if (argv[I][0] == '-')
      {
         if (sscanf (argv[I]+1, "o%s", outName) == 1);
         else if (!strcmp ("n", argv[I]+1)) addComments = 0;
         else if (!strcmp ("?", argv[I]+1)) Help ();
         else
            ErrorExit ("Bad argument: `%s'; type `FSHRINK -?' for help", argv[I]);
      }
      else
         AddPrgm (LoadFargoProgram (argv[I]), &prgms);
   }

   if (!outName[0])
      ErrorExit ("No output file specified; type  `FSHRINK -?' for help");
   if (prgms == NULL)
      ErrorExit ("No input file specified; type  `FSHRINK -?' for help");

   CreatePrgmArchive (prgms, outName, addComments);

   return (0);
}
